/*
    DoubleSimple: purify.c
    Copyright (C) 2003-2006  Macaulay Institute

    This file is part of DoubleSimple, a series of classes for implementing
    various remedies to floating point arithmetic errors.

    DoubleSimple is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    DoubleSimple is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

Collection of functions to use with the various classes that work with numbers
to a specified number of significant figures (DoubleString... & DoubleRound...)
These functions are included in the implementations of each such class to
ensure a standard implementation and lack of duplication.

*/

#import <stdlib.h>
#import <string.h>
#import <errno.h>
#import <math.h>
#ifdef INTEL
#  import <CIieeefp.h>
#endif

double purify_string(double number, int significant_figures) {
  double purified;
  char *buf;
#ifdef INTEL
  fp_pctl save_pctl;

  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  buf = (char *)calloc(significant_figures + 10, sizeof(char));

  if(buf == NULL) {
    perror("Memory allocation");
    abort();
  }

  sprintf(buf, "%.*e", significant_figures - 1, number);
  purified = atof(buf);
  free(buf);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return purified;
}

double purify_string_epsilon(double number,
			    int significant_figures,
			    double epsilon) {
  double purified;

  purified = purify_string(number, significant_figures);

  if(fabs(purified) < epsilon) return 0.0;
  else return purified;
}

double purify_round(double number, int significant_figures) {
  double purified;
  double p10;
#ifdef INTEL
  fp_pctl save_pctl;

  save_pctl = fpsetprecision(FP_PC_DBL);
#endif

  p10 = ceil(log10(fabs(number))) - (double)significant_figures;
  
  purified = (p10 > 0.0)
    ? rint(number / pow(10.0, p10)) * pow(10.0, p10)
    : rint(number * pow(10.0, -p10)) / pow(10.0, -p10);

#ifdef INTEL
  fpsetprecision(save_pctl);
#endif

  return purified;
}

double purify_round_epsilon(double number,
			    int significant_figures,
			    double epsilon) {
  double purified;

  purified = purify_round(number, significant_figures);

  if(fabs(purified) < epsilon) return 0.0;
  else return purified;
}
